#
# Denon AV Receiver control plugin for Indigo 5
#
# Copyright 2012-2017 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyn
import denon
import cyin.devstate
from cyin import iom, plug
from cyin import log, debug, error
from cyin import PluginPreference, DeviceState, PluginProperty
from cyin.asynplugin import action
from cyin.attr import toggle_bool, Toggle
from cyin.check import *

import cyin.eval


class AVRState(cyin.DeviceState):

	def __init__(self, track=None, **kwargs):
		self.track = track
		cyin.DeviceState.__init__(self, **kwargs)


#
# Shared properties of main and zone devices. Both are devstate.Devices.
#
class _Zone(object, metaclass=cyin.devstate.Device.__class__, register=False):

	volume = AVRState(type=float)
	muted = AVRState(type=bool)
	source = AVRState()
	power = AVRState(type=bool)
	quick_select = AVRState(type=int)

	def __init__(self):
		self.zone = None
		# map tracking attribute to state attribute names
		self.attrmap = dict(
			[(attr.track or name, name) for (name, attr) in self.attributes.items()
				if hasattr(attr, 'track')])

	def _event(self, ctx, *args):
		if ctx.state == 'change':
			if self.ready():
				self.changed(*args)
		else:
			debug("DENON EVENT", ctx, *args)

	def changed(self, attr, new_value, old_value):
		""" Respond to change notifications for attribute values. """
		if attr.name in self.attrmap:
			setattr(self, self.attrmap[attr.name], new_value)

	def sync_attributes(self):
		for avrname, myname in self.attrmap.items():
			value = getattr(self.zone, avrname)
			if value:
				setattr(self, myname, value)

	@action
	def set_zone_power(self, action):
		if action.value == Toggle:
			self.zone.power = not self.zone.power
		else:
			self.zone.power = action.value

	@action
	def set_muted(self, action):
		if action.value == Toggle:
			self.zone.muted = not self.zone.muted
		else:
			self.zone.muted = action.value

	@action
	def set_volume(self, action):
		self.zone.volume = action.value

	@action
	def set_source(self, action):
		self.zone.source = action.value

	@action
	def set_quick_select(self, action):
		self.zone.quick_select = action.value

	@action
	def set_attribute(self, action):
		try:
			self.zone.set_attribute(action.name, action.value)
		except KeyError as e:
			error(f'{self.name} has no settable "{action.name}" field')
		except ValueError as e:
			error(str(e))


#
# A receiver and its main zone
#
class AVReceiver(cyin.devstate.IPDevice, cyin.devstate.Idler, _Zone):

	sourcemap = cyin.PluginProperty(reconfigure='notify')

	video_source = AVRState()
	record_source = AVRState()
	surround = AVRState()
	tuner_frequency = AVRState()
	dynamic_volume = AVRState()
	mult_eq = AVRState()
	video_processing = AVRState(track='video_processing_mode')

	def __init__(self, *args):
		_Zone.__init__(self)
		cyin.devstate.IPDevice.__init__(self, *args)
		cyin.devstate.Idler.__init__(self)

	@property
	def avr(self): return self.target

	DEFAULT_PORT = denon.PORT
	SERIAL_CONFIG = dict(baudrate=9600,)

	def connected(self, socket):
		self.target = self.zone = denon.AVR(cyin.plugin, callout=self._event)
		self.proceed("exploring")
		self.avr.open(socket)
		self.avr.discover()
		self.idle_update()

	def config_changed(self, changes):
		if 'sourcemap' in changes:
			self._source_map = None	# recompute
		super(AVReceiver, self).config_changed(changes)

	def _event(self, ctx, *args):
		if ctx.error:
			return self.fail_soft(ctx)
		if ctx.state == 'discover':
			self.discovery(*args)
		elif ctx.state == 'change':
			if self.ready():
				self.changed(*args)
		elif ctx.state in ['send', 'receive']:
			pass
		elif ctx.state == 'END':
			self.reconnect()
		elif ctx.state == 'close':
			pass
		elif ctx.state == 'timeout':
			self.fail_soft("no response from device")
		else:
			_Zone._event(self, ctx, *args)

	@property
	def source_map(self):
		if self._source_map is None:
			def make(s):
				s, _, t = s.partition('=')
				return (s, t if t else s)
			self._source_map = dict([make(s) for s in self.sourcemap.split(',')])
		return self._source_map
	_source_map = None

	def discovery(self, type, *args):
		""" Respond to discovery progress notifications. """
		if type == 'done':
			self.sync_attributes()
			self.proceed("on" if self.avr.main_power else "standby", recovered=True)
		elif type == 'failed':
			self.fail_hard("receiver is not responding")
		elif type == 'zones':
			debug('zones', ', '.join([zone.id for zone in self.avr.zone_list(primary=True)]))

	def changed(self, attr, new_value, old_value):
		""" Respond to change notifications for attribute values. """
		if not self.avr.discovered:	# discovery still ongoing
			return
		if attr.name == 'main_power':
			self.proceed("standby" if not new_value else "on")
		else:
			_Zone.changed(self, attr, new_value, old_value)


	def __getattr__(self, name):
		if name.startswith("set_"):
			return action(lambda action: self.zone.set_attribute(name[4:], action.value))
		raise AttributeError(name)

	@action
	def use_preset(self, action):
		if action.type == "front":
			self.zone.front_preset = int(action.value[1:])
		elif action.type == "tuner":
			self.zone.tuner_preset = action.value
		elif action.type == "net":
			self.zone.network_preset = action.value
		else:
			error('invalid preset type "%s"' % action.type)

	@action
	def menu_navigation(self, action):
		self.avr.command("MN%s" % action.value)


#
# An object for a secondary zone on a receiver.
# This object depends on the primary (main) receiver object.
#
class AVRZone(cyin.devstate.SubDevice, _Zone):

	def __init__(self, *args):
		_Zone.__init__(self)
		cyin.devstate.SubDevice.__init__(self, *args)

	PARTNAME = 'Zone'

	def set_subaddress(self, subaddr):
		self.zoneid = subaddr

	def setup(self, ctx=None):
		avr = self.hostdev.avr
		if self.zoneid not in avr.zones:
			self.fail_hard("no zone %s in %s" % (self.zoneid, self.hostdev.name))
			return
		self.zone = avr.zones[self.zoneid]
		self.zone.add_callout(self._event)
		self.sync_attributes()
		self.proceed("ready")

	@property
	def source_map(self):
		return self.hostdev.source_map


#
# Trigger devices. Can't toggle, can't query, shouldn't read
#
class Trigger(cyin.devstate.Device):

	address = cyin.PluginProperty(type=cyin.device)
	number = cyin.PluginProperty(type=int, check=[check_range(min=1)])

	onOffState = cyin.DeviceState(type=bool)

	def start(self):
		super(Trigger, self).start()
		self.reset()
		self.set_hostdev(self.address)
		self.set_display_address("%s trigger %s" % (self.hostdev.name, self.number))
		if self.hostdev.ready():
			self.setup()

	def setup(self, ctx=None):
		self.avr = self.hostdev.avr
		self.proceed("ready")

	@action
	def standard_switch(self, active):
		debug(self.name, "switched to", active, "by action")
		self.avr.set_trigger(self.number, active)
		self.onOffState = active	# for a while, at least...


#
# Actions
#
class _Setter(cyin.DeviceAction):
	class UI(cyin.ConfigUI):
		def description(self):
			name = self.iomtype.__name__[4:].lower()
			return "set %s of %s to %s" % (name, self.device.name, self.value)

class Set_Main_Power(_Setter):
	value = cyin.PluginProperty(type=bool)

class Set_Zone_Power(_Setter):
	value = cyin.PluginProperty(type=toggle_bool)

class Set_Muted(_Setter):
	value = cyin.PluginProperty(type=toggle_bool)

class Set_Volume(cyin.DeviceAction):
	value = cyin.PluginProperty(type=float, check=[check_range(-100, 20)], eval=True,
		eval_context=lambda self: dict(
			current=self.device.volume,
			step=0.5 if isinstance(self.device, AVReceiver) else 1.0
	))

class Set_Surround(_Setter):
	value = cyin.PluginProperty()

class Set_Source(_Setter):
	value = cyin.PluginProperty(redirect='_other')

class Set_Video_Source(_Setter):
	value = cyin.PluginProperty()

class Set_Record_Source(_Setter):
	value = cyin.PluginProperty()

class Set_Tuner(_Setter):
	value = cyin.PluginProperty(type=float, eval=True,
		eval_context=lambda self: dict(current=self.device.tuner_frequency))

class Set_Mult_EQ(_Setter):
	value = cyin.PluginProperty()

class Set_Dynamic_Volume(_Setter):
	value = cyin.PluginProperty()

class Set_Video_Processing(_Setter):
	value = cyin.PluginProperty()

class Set_Quick_Select(_Setter):
	value = cyin.PluginProperty(type=int, eval=True)

class Use_Preset(_Setter):
	type = cyin.PluginProperty()
	value = cyin.PluginProperty(type=str, eval=True,
		check=[check_format(r'P[123]|[A-G][1-8]', error="Use a two character code such as P2 or C7")])

class Menu_Navigation(_Setter):
	value = cyin.PluginProperty()

class Set_Attribute(cyin.DeviceAction):
	name = cyin.PluginProperty(eval=True)
	value = cyin.PluginProperty(eval=True,
		eval_context=lambda self: dict(current=getattr(self.device.zone, self.name)))
	class UI(cyin.ConfigUI):
		def description(self):
			return "set %s of %s to %s" % (self.name, self.device.name, self.value)

#
# Filters
#
class AVRSourceFilter(cyin.MenuFilter):
	def evaluate(self):
		device = self.ui.device
		if device:
			sources = sorted(device.source_map.items())
			# for zones or secondary sources, add SOURCE at the end
			if isinstance(device, AVRZone) or self.filter == 'secondary':
				sources.append(('SOURCE', '(main)'))
			return sources


class AVRZoneFilter(cyin.MenuGenerator):
	def generate(self):
		for avr in AVReceiver.all():
			for zone in avr.avr.zone_list(primary=False):
				yield (
					"%d@%s" % (avr.id, zone.id),
					"%s Zone %s" % (avr.name, zone.id)
				)


class AVRAttributeFilter(cyin.MenuFilter):
	def evaluate(self):
		device = self.ui.device
		if device and device.zone:
			return sorted([attr.name for attr in device.zone.attributes.values()
				if attr.writable])


#
# The plugin.
#
class Plugin(cyin.asynplugin.Plugin):
	pass
